import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.uci.ics.jung.algorithms.shortestpath.DijkstraShortestPath;
import edu.uci.ics.jung.graph.DirectedEdge;
import edu.uci.ics.jung.graph.Vertex;
import edu.uci.ics.jung.graph.impl.DirectedSparseEdge;
import edu.uci.ics.jung.graph.impl.DirectedSparseGraph;
import edu.uci.ics.jung.graph.impl.SimpleSparseVertex;
import edu.uci.ics.jung.io.GraphMLFile;

import unikl.disco.misc.GraphUtils;
import unikl.disco.nc.Curve;
import unikl.disco.nc.Flow;
import unikl.disco.nc.NetworkAnalyser;
import unikl.disco.nc.TurnProhibition;

/**
 * Simple demo of the capabilities of the DISCO Network Calculator. The code of
 * the demo is hopefully self-explanatory.
 * 
 * @author Frank A. Zdarsky
 */
public class DNCLibDemo {

	private String filename;

	private DirectedSparseGraph network_graph;

	private DirectedSparseGraph server_graph;

	private NetworkAnalyser na;

	public DNCLibDemo(String filename) {
		this.filename = filename;
	}

	public DirectedSparseGraph createDemoNetwork() {
		DirectedSparseGraph network_graph = new DirectedSparseGraph();
		for (int i = 0; i < 11; i++) {
			network_graph.addVertex(new SimpleSparseVertex());
		}
		Vertex[] vertices = (Vertex[]) network_graph.getVertices().toArray(
				new Vertex[network_graph.numVertices()]);

		network_graph.addEdge(new DirectedSparseEdge(vertices[0], vertices[1]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[0], vertices[2]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[0], vertices[3]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[1], vertices[3]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[2], vertices[3]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[2], vertices[4]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[3], vertices[5]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[4], vertices[5]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[5], vertices[6]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[6], vertices[7]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[7], vertices[8]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[7], vertices[10]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[8], vertices[9]));
		network_graph.addEdge(new DirectedSparseEdge(vertices[9], vertices[10]));

		// make graph bidirected
		Set edges = new HashSet(network_graph.getEdges());
		for (Iterator iter = edges.iterator(); iter.hasNext();) {
			DirectedEdge e = (DirectedEdge) iter.next();
			network_graph.addEdge(new DirectedSparseEdge(e.getDest(), e.getSource()));
		}
		return network_graph;
	}

	public DirectedSparseGraph convertFromNetworkToServerGraph(
			DirectedSparseGraph network_graph) {
		DirectedSparseGraph server_graph;

		Map link_server_map = new HashMap();
		Map router_turns_map = new HashMap();
		server_graph = GraphUtils.createServerGraph(network_graph,
				link_server_map, router_turns_map);
		(new GraphMLFile()).save(server_graph, filename + "_server.graphml");

		TurnProhibition tp = new TurnProhibition(network_graph);
		tp.runTurnProhibition();
		tp.removeProhibitedTurns(server_graph, link_server_map,
				router_turns_map);
		(new GraphMLFile()).save(server_graph, filename + "_server_cyclefree.graphml");

		return server_graph;
	}

	public void run() {
		// create a network graph manually...
		network_graph = createDemoNetwork();
		(new GraphMLFile()).save(network_graph, filename + "_network.graphml");
		// ... or by loading an existing graph, e.g. created and exported using BRITE
//		network_graph = (DirectedSparseGraph) (new GraphMLFile()).load(filename + "_network.graphml");

		// convert this graph to a server graph
		server_graph = convertFromNetworkToServerGraph(network_graph);

		// create a new network analyser instance for the server graph
		na = new NetworkAnalyser(server_graph);

		// create a service curve for each server
		Set servers = GraphUtils.getServerSet(server_graph);
		na.setServiceCurve(servers, Curve.createRateLatency(0.01, 10.0e6));
		na.setMaxServiceCurve(servers, Curve.createRateLatency(0.001, 100.0e6));

		// create a flow between every source-sink-pair if a path exists
		DijkstraShortestPath shortest_paths = new DijkstraShortestPath(server_graph);
		Curve flow_prototype = Curve.createTokenBucket(0.1 * 0.1e6, 0.1e6);
		Set sources = GraphUtils.getSourceSet(server_graph);
		Set sinks = GraphUtils.getSinkSet(server_graph);
		for (Iterator sourceIter = sources.iterator(); sourceIter.hasNext();) {
			Vertex source = (Vertex) sourceIter.next();
			for (Iterator sinkIter = sinks.iterator(); sinkIter.hasNext();) {
				Vertex sink = (Vertex) sinkIter.next();

				List path = shortest_paths.getPath(source, sink);
				if (path.size() > 0) {
					na.addFlow(new Flow(source, sink, flow_prototype.copy(), path));
				}
			}
		}

		// select the flow of interest
		Flow flow_of_interest = (Flow) na.getFlows().get(0);
		System.out.println("Flow of interest : " + flow_of_interest.toString());
		System.out.println();

		// analyse the network
		Curve beta;
		Map server_bound_map;

		na.setUseFifoService(false);
		na.setUseGamma(false);
		na.setUseExtraGamma(false);

		System.out.println("--- Charny Bound ---");
		System.out.println("cur. utilization: "	+ na.getHighestServerUtilization());
		System.out.println("max. utilization: " + na.getMaxUtilizationForCharnyBound());
		System.out.println("delay bound     : " + na.getCharnyDelayBound());
		server_bound_map = na.getCharnyBacklogBounds();
		if (!server_bound_map.isEmpty()) {
			System.out.println("backlog bound   : "	+ Collections.max(server_bound_map.values()));
			System.out.println("     per server : " + server_bound_map.values().toString());
		} else {
			System.out.println("backlog bound   : " + Double.POSITIVE_INFINITY);
		}
		System.out.println();

		System.out.println("--- Fair Queueing ---");
		beta = na.performFairQueueingAnalysis(flow_of_interest);
		System.out.println("delay bound     : "	+ Curve.getDelayBound(flow_of_interest.ac, beta));
		System.out.println("backlog bound   : "	+ Curve.getBacklogBound(flow_of_interest.ac, beta));
		System.out.println();

		System.out.println("--- Total Flow Analysis ---");
		server_bound_map.clear();
		double delay = na.performTotalFlowAnalysis(flow_of_interest,
				server_bound_map);
		System.out.println("delay bound     : " + delay);
		System.out.println("backlog bound   : "	+ Collections.max(server_bound_map.values()));
		System.out.println("     per server : " + server_bound_map.values().toString());
		System.out.println();

		System.out.println("--- Separated Flow Analysis ---");
		beta = na.performSeparatedFlowAnalysis(flow_of_interest);
		System.out.println("delay bound     : "	+ Curve.getDelayBound(flow_of_interest.ac, beta));
		System.out.println("backlog bound   : "	+ Curve.getBacklogBound(flow_of_interest.ac, beta));
		System.out.println();

		System.out.println("--- PMOO Analysis ---");
		beta = na.performPMOOAnalysis(flow_of_interest);
		System.out.println("delay bound     : "	+ Curve.getDelayBound(flow_of_interest.ac, beta));
		System.out.println("backlog bound   : "	+ Curve.getBacklogBound(flow_of_interest.ac, beta));
		System.out.println();
	}

	public static void main(String[] args) {
		DNCLibDemo demo = new DNCLibDemo("DNCLibDemo");
		demo.run();
	}
}
